import os
import signal
import threading
import time
from contextlib import nullcontext
from multiprocessing import Pipe
from multiprocessing.context import SpawnProcess
from typing import Any, Callable, ParamSpec

from .logger import logger

UNIX_SIGNALS = {
    getattr(signal, f"SIG{x}"): x
    for x in "HUP QUIT TTIN TTOU USR1 USR2 WINCH".split()
    if hasattr(signal, f"SIG{x}")
}

P = ParamSpec("P")


class ProcessParameters:
    def __init__(
        self,
        f: Callable[P, Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.f = f
        self.args = args
        self.kwargs = kwargs


class Process:
    def __init__(self, parameters: ProcessParameters) -> None:
        self.parameters = parameters
        self.parent_conn, self.child_conn = Pipe()
        self.process = SpawnProcess(target=self.target, daemon=True)

    def ping(self, timeout: float = 5) -> bool:
        try:
            self.parent_conn.send(b"ping")
            if self.parent_conn.poll(timeout):
                self.parent_conn.recv()
                return True
        except IOError:
            pass
        return False

    def pong(self) -> None:
        self.child_conn.recv()
        self.child_conn.send(b"pong")

    def always_pong(self) -> None:
        while True:
            self.pong()

    def target(self) -> Any:
        if os.name == "nt":
            # Windows doesn't support SIGTERM, so we use SIGBREAK instead.
            # And then we raise SIGTERM when SIGBREAK is received.
            # https://learn.microsoft.com/zh-cn/cpp/c-runtime-library/reference/signal?view=msvc-170
            signal.signal(
                signal.SIGBREAK, lambda sig, frame: signal.raise_signal(signal.SIGTERM)
            )

        threading.Thread(target=self.always_pong, daemon=True).start()
        return self.parameters.f(*self.parameters.args, **self.parameters.kwargs)

    def is_alive(self, timeout: float = 5) -> bool:
        if not self.process.is_alive():
            return False

        return self.ping(timeout)

    def start(self) -> None:
        self.process.start()
        logger.info("Started child process [{}]".format(self.process.pid))

    def terminate(self) -> None:
        if self.process.exitcode is not None:
            return
        assert self.process.pid is not None
        if os.name == "nt":
            # Windows doesn't support SIGTERM.
            # So send SIGBREAK, and then in process raise SIGTERM.
            os.kill(self.process.pid, signal.CTRL_BREAK_EVENT)
        else:
            os.kill(self.process.pid, signal.SIGTERM)
        logger.info("Terminated child process [{}]".format(self.process.pid))

        self.parent_conn.close()
        self.child_conn.close()

    def terminate_quickly(self) -> None:
        if self.process.exitcode is not None:
            return
        assert self.process.pid is not None
        if os.name == "nt":
            os.kill(self.process.pid, signal.CTRL_C_EVENT)
        else:
            os.kill(self.process.pid, signal.SIGINT)
        logger.info("Terminated quickly child process [{}]".format(self.process.pid))

        self.parent_conn.close()
        self.child_conn.close()

    def kill(self) -> None:
        # In Windows, the method will call `TerminateProcess` to kill the process.
        # In Unix, the method will send SIGKILL to the process.
        self.process.kill()

    def join(self, timeout: float | None = None) -> None:
        logger.info("Waiting for child process [{}]".format(self.process.pid))
        self.process.join(timeout)
        # Timeout, kill the process
        while self.process.exitcode is None:
            self.process.kill()
            self.process.join(1)

    @property
    def pid(self) -> int | None:
        return self.process.pid


class MultiProcessManager:
    def __init__(self, processes_num: int, process_parameters: ProcessParameters):
        self.processes_num = processes_num
        self.process_parameters = process_parameters
        self.processes: list[Process] = []

        self.should_exit = threading.Event()
        self.reloading = False

        self.signal_queue: list[int] = []
        for sig in UNIX_SIGNALS:
            signal.signal(sig, lambda sig, frame: self.signal_queue.append(sig))

        # Sent by Ctrl+C.
        signal.signal(signal.SIGINT, lambda sig, frame: self.handle_int())
        # Sent by `kill <pid>`. Not sent on Windows.
        signal.signal(signal.SIGTERM, lambda sig, frame: self.handle_term())
        if os.name == "nt":
            # Sent by `Ctrl+Break` on Windows.
            signal.signal(signal.SIGBREAK, lambda sig, frame: self.handle_break())

    def init_processes(self) -> None:
        for _ in range(self.processes_num):
            process = Process(self.process_parameters)
            process.start()
            self.processes.append(process)

    def terminate_all(self) -> None:
        for process in self.processes:
            process.terminate()

    def terminate_all_quickly(self) -> None:
        for process in self.processes:
            process.terminate_quickly()

    def join_all(self, timeout: float | None = None) -> None:
        for process in self.processes:
            process.join(timeout)

    def restart_all(self) -> None:
        for idx, process in enumerate(tuple(self.processes)):
            process.terminate()
            process.join()
            del self.processes[idx]
            process = Process(self.process_parameters)
            process.start()
            self.processes.append(process)

    def on_watchfiles_reload(self) -> None:
        self.reloading = True
        self.terminate_all_quickly()
        self.join_all()
        time.sleep(1)  # Wait for the Ctrl+C signal to be handled
        # Because in Windows, the Ctrl+C signal always send to main process.
        self.reloading = False

    def mainloop(self) -> None:
        logger.info("Started parent process [{}]".format(os.getpid()))

        self.init_processes()

        while not self.should_exit.wait(0.5):
            self.handle_signals()
            self.keep_subprocess_alive()

        self.join_all()

        logger.info("Stopped parent process [{}]".format(os.getpid()))

    def keep_subprocess_alive(self) -> None:
        if self.should_exit.is_set():
            return  # parent process is exiting, no need to keep subprocess alive

        for idx, process in enumerate(tuple(self.processes)):
            if process.is_alive():
                continue

            process.kill()  # process is hung, kill it
            process.join(1)

            if self.should_exit.is_set():
                return

            logger.info("Child process [{}] died".format(process.pid))
            del self.processes[idx]
            process = Process(self.process_parameters)
            process.start()
            self.processes.append(process)

    def handle_signals(self) -> None:
        for sig in tuple(self.signal_queue):
            self.signal_queue.remove(sig)
            sig_name = UNIX_SIGNALS[sig]
            sig_handler = getattr(self, f"handle_{sig_name.lower()}", None)
            if sig_handler is not None:
                sig_handler()
            else:
                logger.info(f"Received signal [{sig_name}], but nothing to do")

    def handle_int(self) -> None:
        if self.reloading:
            return
        logger.info("Received SIGINT, quickly exiting")
        self.should_exit.set()
        self.terminate_all_quickly()

    def handle_term(self) -> None:
        logger.info("Received SIGTERM, exiting")
        self.should_exit.set()
        self.terminate_all()

    def handle_break(self) -> None:
        if self.reloading:
            return
        logger.info("Received SIGBREAK, exiting")
        self.should_exit.set()
        self.terminate_all()

    def handle_hup(self) -> None:
        logger.info("Received SIGHUP, restarting processes")
        self.restart_all()

    def handle_ttin(self) -> None:
        logger.info("Received SIGTTIN, increasing processes")
        self.processes_num += 1
        process = Process(self.process_parameters)
        process.start()
        self.processes.append(process)

    def handle_ttou(self) -> None:
        logger.info("Received SIGTTOU, decreasing processes")
        if self.processes_num <= 1:
            logger.info("Cannot decrease processes any more")
            return
        self.processes_num -= 1
        process = self.processes.pop()
        process.terminate()
        process.join()


def multiprocess(
    processes_num: int,
    process_parameters: ProcessParameters,
    watchfiles: str | None,
) -> None:
    processes_manager = MultiProcessManager(processes_num, process_parameters)

    if watchfiles is not None:
        from .reloader import listen_for_changes

        contextmanager = listen_for_changes(
            watchfiles, processes_manager.on_watchfiles_reload
        )
    else:
        contextmanager = nullcontext()

    with contextmanager:
        processes_manager.mainloop()



import os
import random

from zibai import create_bind_socket


def create_ipv4_socket():
    for _ in range(10):
        port = random.randint(20000, 60000)
        try:
            return create_bind_socket(f"127.0.0.1:{port}")
        except IOError:
            continue


def create_ipv6_socket():
    for _ in range(10):
        port = random.randint(20000, 60000)
        try:
            return create_bind_socket(f"::1:{port}")
        except IOError:
            continue


def create_unix_socket():
    path = "/tmp/test-zibai-server.sock"
    if os.path.exists(path):
        os.remove(path)
    return create_bind_socket(f"unix:{path}")



import socket
import threading

import pytest

from .utils import create_ipv4_socket, create_ipv6_socket, create_unix_socket


@pytest.fixture(params=[create_ipv4_socket, create_ipv6_socket, create_unix_socket])
def bind_socket(request):
    if request.param is create_unix_socket:
        if not hasattr(socket, "AF_UNIX"):
            pytest.skip("AF_UNIX is not supported")

    with request.param() as sock:
        yield sock


@pytest.fixture
def exit_event():
    exit_event = threading.Event()
    try:
        yield exit_event
    finally:
        exit_event.set()



import argparse
import dataclasses
import importlib
import ipaddress
import logging
import multiprocessing
import os
import signal
import socket
import sys
import threading
from functools import reduce
from typing import Any, Sequence

from .core import serve
from .logger import logger
from .multiprocess import ProcessParameters, multiprocess
from .wsgi_typing import WSGIApp


@dataclasses.dataclass
class Options:
    """
    Keep `Options` can be passed between processes.
    """

    app: str
    call: bool = False
    listen: str = "127.0.0.1:9000"
    subprocess: int = 0
    no_gevent: bool = False
    max_workers: int = 10
    watchfiles: str | None = None

    backlog: int | None = None
    dualstack_ipv6: bool = False
    unix_socket_perms: int = 0o600
    h11_max_incomplete_event_size: int | None = None
    max_request_pre_process: int | None = None
    graceful_exit_timeout: float = 10

    # WSGI environment settings
    url_scheme: str = "http"
    url_prefix: str | None = None

    # Server callback hooks
    before_serve: str | None = None
    before_graceful_exit: str | None = None
    before_died: str | None = None

    # Logging
    no_access_log: bool = False

    def __post_init__(self) -> None:
        """
        Check options. Do not do any side effects here.
        """
        if self.watchfiles is not None and self.subprocess <= 0:
            raise ValueError("Cannot watch files without subprocesses")

        if self.dualstack_ipv6 and not socket.has_dualstack_ipv6():
            raise ValueError("Dualstack ipv6 is not supported on this platform")

    @classmethod
    def default_value(cls, field_name: str) -> Any:
        fields = {field.name: field for field in dataclasses.fields(Options)}
        default = fields[field_name].default
        if default is dataclasses.MISSING:
            raise ValueError(f"Field {field_name} has no default value")
        return default


def import_from_string(import_str: str) -> Any:
    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        raise ValueError(
            f'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )

    return reduce(getattr, attrs_str.split("."), importlib.import_module(module_str))


def create_bind_socket(
    value: str,
    *,
    uds_perms: int = Options.default_value("unix_socket_perms"),
    dualstack_ipv6: bool = Options.default_value("dualstack_ipv6"),
    socket_type: int = socket.SOCK_STREAM,
) -> socket.socket:
    if value.startswith("unix:"):
        if not hasattr(socket, "AF_UNIX"):
            raise ValueError("UNIX sockets are not supported on this platform")

        path = value[5:]
        sock = socket.socket(socket.AF_UNIX, socket_type)  # type: ignore
        sock.bind(path)

        os.chmod(path, uds_perms)
        return sock

    if ":" not in value:
        raise ValueError("Bind must be of the form: HOST:PORT")

    host, port = value.rsplit(":", 1)

    try:
        port = int(port)
    except ValueError:
        raise ValueError("Bind port must be an integer")

    if not 0 < port < 65536:
        raise ValueError("Bind port must be between 0 and 65536")

    if host == "":
        if dualstack_ipv6:
            host = "::"
        else:
            host = "0.0.0.0"

    address = ipaddress.ip_address(host)
    sock = socket.socket(
        socket.AF_INET if address.version == 4 else socket.AF_INET6,
        socket_type,
    )

    # Set socket options
    if dualstack_ipv6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
    if os.name != "nt":
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    else:  # In windows, SO_REUSEPORT is not available
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # Bind socket
    sock.bind((str(address), port))

    return sock


def parse_args(args: Sequence[str]) -> Options:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    # Please keep the order of arguments like `Options`.
    parser.add_argument("app", help="WSGI app")
    parser.add_argument(
        "--call",
        help="use WSGI factory",
        default=Options.default_value("call"),
        action="store_true",
    )
    parser.add_argument(
        "--listen",
        "-l",
        default=Options.default_value("listen"),
        help="listen address, HOST:PORT, unix:PATH",
    )
    parser.add_argument(
        "--subprocess",
        "-p",
        default=Options.default_value("subprocess"),
        type=int,
        help="number of subprocesses",
    )
    parser.add_argument(
        "--no-gevent",
        default=Options.default_value("no_gevent"),
        action="store_true",
        help="do not use gevent",
    )
    parser.add_argument(
        "--max-workers",
        "-w",
        default=Options.default_value("max_workers"),
        type=int,
        help="maximum number of threads or greenlets to use for handling requests",
    )
    parser.add_argument(
        "--watchfiles",
        help="watch files for changes and restart workers",
        required=False,
    )
    parser.add_argument(
        "--backlog",
        type=int,
        help="listen backlog",
        required=False,
    )
    parser.add_argument(
        "--dualstack-ipv6",
        default=Options.default_value("dualstack_ipv6"),
        action="store_true",
        help="enable dualstack ipv6",
    )
    parser.add_argument(
        "--unix-socket-perms",
        default="600",
        help="unix socket permissions",
    )
    parser.add_argument(
        "--h11-max-incomplete-event-size",
        type=int,
        help="maximum number of bytes in an incomplete HTTP event",
        required=False,
    )
    parser.add_argument(
        "--max-request-pre-process",
        type=int,
        help="maximum number of requests to process before killing the worker",
        required=False,
    )
    parser.add_argument(
        "--graceful-exit-timeout",
        default=Options.default_value("graceful_exit_timeout"),
        type=float,
        help="graceful exit timeout",
    )
    parser.add_argument(
        "--url-scheme",
        default=Options.default_value("url_scheme"),
        help="url scheme; will be passed to WSGI app as wsgi.url_scheme",
    )
    parser.add_argument(
        "--url-prefix",
        help="url prefix; will be passed to WSGI app as SCRIPT_NAME, "
        "if not specified, use environment variable SCRIPT_NAME",
        required=False,
    )
    parser.add_argument(
        "--before-serve",
        help="callback to run before serving requests",
        required=False,
    )
    parser.add_argument(
        "--before-graceful-exit",
        help="callback to run before graceful exit",
        required=False,
    )
    parser.add_argument(
        "--before-died",
        help="callback to run before exiting",
        required=False,
    )
    parser.add_argument(
        "--no-access-log",
        default=Options.default_value("no_access_log"),
        action="store_true",
        help="disable access log",
    )
    options = parser.parse_args(args)

    # Parse unix_socket_perms as an octal integer.
    options.unix_socket_perms = int(options.unix_socket_perms, base=8)

    # When watchfiles is specified, subprocess must be greater than 0.
    if options.watchfiles is not None:
        options.subprocess = max(options.subprocess, 1)

    return Options(**options.__dict__)


spawn = multiprocessing.get_context("spawn")


def get_app(string: str, use_factory: bool = False) -> Any:
    """
    Get WSGI app from import string.
    """
    app = import_from_string(string)
    if use_factory:
        app = app()
    return app


def main(options: Options, *, is_main: bool = True) -> None:
    if not options.no_gevent and (options.subprocess == 0 or not is_main):
        # Single process mode or worker process with gevent.
        try:
            import gevent
        except ImportError:
            logger.warning("gevent not found, using threading instead")
        else:
            import gevent.monkey

            gevent.monkey.patch_all()
            logger.info("Using gevent for worker pool")

    if options.no_access_log:
        logging.getLogger("zibai.access").setLevel(logging.WARNING)

    # Check that app is importable.
    get_app(options.app, use_factory=options.call)
    # Check that bind socket can be created.
    create_bind_socket(
        options.listen,
        uds_perms=options.unix_socket_perms,
        dualstack_ipv6=options.dualstack_ipv6,
    ).close()

    if is_main and options.subprocess > 0:
        multiprocess(
            options.subprocess,
            ProcessParameters(main, options, is_main=False),
            options.watchfiles,
        )
        return

    if options.h11_max_incomplete_event_size is not None:
        # Set max_incomplete_event_size
        from . import h11

        h11.MAX_INCOMPLETE_EVENT_SIZE = options.h11_max_incomplete_event_size

    if options.before_serve is not None:
        before_serve_hook = import_from_string(options.before_serve)
    else:
        before_serve_hook = lambda: None

    if options.before_graceful_exit is not None:
        before_graceful_exit_hook = import_from_string(options.before_graceful_exit)
    else:
        before_graceful_exit_hook = lambda: None

    if options.before_died is not None:
        before_died_hook = import_from_string(options.before_died)
    else:
        before_died_hook = lambda: None

    application: WSGIApp = import_from_string(options.app)
    if options.max_request_pre_process is not None:
        from .middlewares.limit_request_count import LimitRequestCountMiddleware

        application = LimitRequestCountMiddleware(
            application, options.max_request_pre_process
        )

    graceful_exit = threading.Event()

    def handle_int(sig, frame) -> None:
        logger.info("Received SIGINT, qucikly exiting")
        graceful_exit.set()
        sys.exit(0)

    def handle_term(sig, frame) -> None:
        if graceful_exit.is_set():
            logger.info("Received second SIGTERM, quickly exiting")
            sys.exit(0)
        logger.info("Received SIGTERM, gracefully exiting")
        graceful_exit.set()

    signal.signal(signal.SIGINT, handle_int)
    signal.signal(signal.SIGTERM, handle_term)

    if is_main:
        logger.info("Run in single process mode [%d]", os.getpid())

    serve(
        app=application,
        bind_socket=create_bind_socket(
            options.listen,
            uds_perms=options.unix_socket_perms,
            dualstack_ipv6=options.dualstack_ipv6,
        ),
        backlog=options.backlog,
        max_workers=options.max_workers,
        graceful_exit=graceful_exit,
        graceful_exit_timeout=options.graceful_exit_timeout,
        before_serve_hook=before_serve_hook,
        before_graceful_exit_hook=before_graceful_exit_hook,
        before_died_hook=before_died_hook,
    )



import dataclasses
import signal

from ..wsgi_typing import Environ, IterableChunks, StartResponse, WSGIApp


@dataclasses.dataclass
class LimitRequestCountMiddleware:
    app: WSGIApp
    max_request_pre_process: int
    request_count: int = 0

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> IterableChunks:
        yield from self.app(environ, start_response)

        self.request_count += 1
        if self.request_count >= self.max_request_pre_process:
            signal.raise_signal(signal.SIGTERM)



import os
import threading
from contextlib import contextmanager
from typing import Any, Callable, Generator

import watchdog.events
from watchdog.observers import Observer

from .logger import logger


@contextmanager
def listen_for_changes(
    watchfiles: str, callback: Callable[[], Any]
) -> Generator[None, None, None]:
    """
    When any of the files in `watchfiles` matches a change, call `callback`.
    """
    reloading_event = threading.Event()

    def run_callback(event: watchdog.events.FileSystemEvent) -> None:
        if not reloading_event.is_set():
            logger.info("Detected file change, reloading...")
            reloading_event.set()
            callback()
            reloading_event.clear()

    def on_any_event(event: watchdog.events.FileSystemEvent) -> None:
        if event.event_type not in (
            watchdog.events.EVENT_TYPE_MOVED,
            watchdog.events.EVENT_TYPE_DELETED,
            watchdog.events.EVENT_TYPE_CREATED,
            watchdog.events.EVENT_TYPE_MODIFIED,
        ):
            return

        threading.Thread(target=run_callback, args=(event,), daemon=True).start()

    event_handler = watchdog.events.PatternMatchingEventHandler(
        patterns=[pattern.strip() for pattern in watchfiles.split(";")],
    )

    setattr(event_handler, "on_any_event", on_any_event)

    observer = Observer()

    path = os.getcwd()
    logger.info("Watching files in {}".format(path))
    observer.schedule(event_handler, path, recursive=True)

    observer.start()
    try:
        logger.debug("Started observer")
        yield
    finally:
        logger.debug("Stopping observer")
        observer.stop()
        observer.join()



import os
import socket
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any, Callable, Generator

from .h11 import http11_protocol
from .logger import debug_logger, logger
from .utils import unicode_to_wsgi
from .wsgi_typing import WSGIApp


def handle_connection(
    app: WSGIApp,
    s: socket.socket,
    address: tuple[str, int],
    graceful_exit: threading.Event,
    *,
    url_scheme: str = "http",
    script_name: str = "",
) -> None:
    debug_logger.debug("Handling connection from %s:%d", *address[:2])
    with s:
        try:
            http11_protocol(
                app,
                s,
                graceful_exit,
                url_scheme=url_scheme,
                script_name=script_name,
            )
        except ConnectionError:
            pass  # client closed connection, nothing to do


@contextmanager
def lifespan_hooks_context(
    before_serve_hook: Callable[[], None] = lambda: None,
    before_died_hook: Callable[[], None] = lambda: None,
) -> Generator[None, None, None]:
    """
    Context manager for lifespan hooks.
    """
    before_serve_hook()
    try:
        yield
    finally:
        before_died_hook()


def serve(
    *,
    app: Any,
    bind_socket: socket.socket,
    max_workers: int,
    graceful_exit: threading.Event,
    graceful_exit_timeout: float = 10,
    backlog: int | None = None,
    url_scheme: str = "http",
    script_name: str | None = None,
    before_serve_hook: Callable[[], None] = lambda: None,
    before_graceful_exit_hook: Callable[[], None] = lambda: None,
    before_died_hook: Callable[[], None] = lambda: None,
) -> None:
    """
    Serve a WSGI application.
    """
    if script_name is None:
        # If script_name is not specified, use the environment variable.
        script_name = unicode_to_wsgi(os.environ.get("SCRIPT_NAME", ""))

    listen_address = bind_socket.getsockname()[:2]

    def _handle_exit_event() -> None:
        graceful_exit.wait()
        try:
            before_graceful_exit_hook()
        except Exception:  # pragma: no cover
            logger.exception("Exception in `before_graceful_exit` callback")
        finally:
            bind_socket.close()
            logger.info("Stopped listening on %s:%d", *listen_address)
            time.sleep(graceful_exit_timeout)
            logger.info("Graceful exit timeout, force exit")
            os.kill(os.getpid(), 9)

    threading.Thread(target=_handle_exit_event, daemon=True).start()

    lifespan_hooks = lifespan_hooks_context(
        before_serve_hook=before_serve_hook, before_died_hook=before_died_hook
    )
    executor = ThreadPoolExecutor(
        max_workers=max_workers, thread_name_prefix="zibai_worker"
    )

    with lifespan_hooks, bind_socket, executor:
        if backlog is not None:
            bind_socket.listen(backlog)
        else:
            bind_socket.listen()
        logger.info("Accepting request on %s:%d", *listen_address)

        while not graceful_exit.is_set():
            try:
                connection, address = bind_socket.accept()
            except OSError:  # bind_socket closed
                if not graceful_exit.is_set():
                    raise  # pragma: no cover
            else:
                debug_logger.debug("Accepted connection from %s:%d", *address[:2])
                future = executor.submit(
                    handle_connection,
                    app,
                    connection,
                    address,
                    graceful_exit,
                    url_scheme=url_scheme,
                    script_name=script_name,
                )
                # raise exception in main thread if exception in threadpool
                future.add_done_callback(lambda future: future.result())



import logging

# Set default logging format.
logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")

logger = logging.getLogger("zibai")

debug_logger = logging.getLogger("zibai.debug")
debug_logger.setLevel(logging.INFO)  # Disable debug logging by default.

access_logger = logging.getLogger("zibai.access")

error_logger = logging.getLogger("zibai.error")


def log_http(environ, status_code) -> None:
    if status_code >= 500:
        error_logger.error(
            '"%s %s %s" %s',
            environ["REQUEST_METHOD"],
            environ["PATH_INFO"],
            environ["SERVER_PROTOCOL"],
            status_code,
            extra=environ,
            exc_info=True,
        )
    else:
        access_logger.info(
            '"%s %s %s" %s',
            environ["REQUEST_METHOD"],
            environ["PATH_INFO"],
            environ["SERVER_PROTOCOL"],
            status_code,
            extra=environ,
        )



from .cli import (
    Options,
    create_bind_socket,
    import_from_string,
    main,
    parse_args,
)
from .core import serve
from .multiprocess import MultiProcessManager, ProcessParameters

__all__ = [
    "parse_args",
    "Options",
    "main",
    "create_bind_socket",
    "import_from_string",
    "serve",
    "MultiProcessManager",
    "ProcessParameters",
]



import dataclasses
import socket
import sys
import threading
from typing import Any, Callable

import h11

from .const import SERVER_NAME
from .logger import debug_logger, error_logger, log_http
from .utils import Input
from .wsgi_typing import Environ, ExceptionInfo, WSGIApp


class ConnectionClosed(Exception):
    """
    Received a ConnectionClosed event from h11.
    """


@dataclasses.dataclass
class H11Protocol:
    c: h11.Connection
    s: socket.socket
    peername: tuple[str, int]
    sockname: tuple[str, int]

    graceful_exit: threading.Event

    # WSGI variables
    url_scheme: str = "http"
    script_name: str = ""

    # State variables
    response_buffer: tuple[int, list[tuple[bytes, bytes]]] | None = None
    header_sent: bool = False

    def __post_init__(self):
        self.s.settimeout(1)  # For graceful exit

    def get_next_event(self):
        if self.c.their_state is h11.DONE:
            return h11.PAUSED

        while True:
            if self.c.their_state is h11.IDLE and self.graceful_exit.is_set():
                raise ConnectionClosed

            event = self.c.next_event()
            debug_logger.debug("Received event from %s:%d: %r", *self.peername, event)

            match event:
                case h11.NEED_DATA:
                    if self.c.they_are_waiting_for_100_continue:
                        self.send_with_event(
                            h11.InformationalResponse(headers=[], status_code=100)
                        )
                    try:
                        self.c.receive_data(self.s.recv(MAX_INCOMPLETE_EVENT_SIZE))
                    except socket.timeout:
                        pass
                case h11.ConnectionClosed():
                    debug_logger.debug("Connection closed by %s:%d", *self.peername)
                    raise ConnectionClosed
                case _:
                    return event

    def send_with_event(self, event) -> None:
        data = self.c.send(event)
        assert data is not None
        self.s.sendall(data)
        debug_logger.debug("Sent event to %s:%d: %r", *self.peername, event)

    def read_request_body(self):
        event = self.get_next_event()
        match event:
            case h11.Data(data):
                return data
            case h11.EndOfMessage():
                return b""
            case _:
                return b""

    def start_response(
        self,
        status: str,
        headers: list[tuple[str, str]],
        exc_info: ExceptionInfo | None = None,
    ) -> Callable[[bytes], Any]:
        if exc_info is not None:
            try:
                if self.header_sent:
                    raise exc_info[1].with_traceback(exc_info[2])
            finally:
                del exc_info
        elif self.response_buffer is not None:
            raise RuntimeError("start_response() was already called")

        status_code, _ = status.split(" ", 1)
        if status_code.isdigit():
            status_code = int(status_code)
        else:
            raise RuntimeError(f"Invalid status: {status}")

        self.response_buffer = (
            status_code,
            [
                *(
                    (name.encode("latin1"), value.encode("latin1"))
                    for name, value in headers
                ),
                (b"Server", SERVER_NAME),
            ],
        )

        return self.s.sendall

    def init_environ(self) -> Environ:
        event = self.get_next_event()
        match event:
            case h11.Request(method, headers, target, http_version):
                request_uri = target.decode("latin-1")
                if "?" in request_uri:
                    path, query = request_uri.split("?", 1)
                else:
                    path, query = request_uri, ""

                server_name, server_port = self.sockname
                remote_name, remote_port = self.peername

                script_name = self.script_name
                if path == script_name:
                    path = ""
                else:
                    url_prefix_with_trailing_slash = script_name + "/"
                    if path.startswith(url_prefix_with_trailing_slash):
                        path = path[len(script_name) :]

                environ: Environ = {
                    "REQUEST_METHOD": method.decode("ascii"),
                    "SCRIPT_NAME": script_name,
                    "SERVER_NAME": server_name,
                    "SERVER_PORT": str(server_port),
                    "REMOTE_ADDR": remote_name,
                    "REMOTE_PORT": str(remote_port),
                    "REQUEST_URI": request_uri,
                    "PATH_INFO": path,
                    "QUERY_STRING": query,
                    "SERVER_PROTOCOL": f"HTTP/{http_version.decode('ascii')}",
                    "wsgi.version": (1, 0),
                    "wsgi.url_scheme": self.url_scheme,
                    "wsgi.input": Input(self.read_request_body),
                    "wsgi.errors": sys.stderr,
                    "wsgi.multithread": True,
                    "wsgi.multiprocess": True,
                    "wsgi.run_once": False,
                }

                for name, value in headers:
                    name = name.decode("latin1")
                    value = value.decode("latin1")
                    if name == "content-type":
                        environ["CONTENT_TYPE"] = value
                    elif name == "content-length":
                        environ["CONTENT_LENGTH"] = value
                    else:
                        http_name = "HTTP_" + name.upper().replace("-", "_")
                        if http_name not in environ:
                            environ[http_name] = value
                        else:
                            environ[http_name] += "," + value

                return environ  # type: ignore
            case _:
                raise RuntimeError(f"Unexpected event: {event}")

    def call_wsgi(self, wsgi_app: WSGIApp) -> None:
        environ = self.init_environ()
        iterable = None  # Just for finally block

        try:
            iterable = wsgi_app(environ, self.start_response)
            iterator = iter(iterable)

            chunk = next(iterator)
            if self.response_buffer is None:
                raise RuntimeError("start_response() was not called")

            status_code = self.response_buffer[0]
            self.send_with_event(
                h11.Response(status_code=status_code, headers=self.response_buffer[1])
            )
            self.header_sent = True

            log_http(environ, status_code)

            self.send_with_event(h11.Data(data=chunk))

            for chunk in iterator:
                self.send_with_event(h11.Data(data=chunk))

            self.send_with_event(h11.EndOfMessage())
        except Exception:
            if self.header_sent:
                raise

            self.send_with_event(
                h11.Response(
                    status_code=500,
                    headers=[
                        (b"Content-Type", b"text/plain; charset=utf-8"),
                        (b"Content-Length", b"21"),
                        (b"Server", SERVER_NAME),
                    ],
                )
            )
            self.send_with_event(h11.Data(data=b"Internal Server Error"))
            self.send_with_event(h11.EndOfMessage())

            error_logger.exception(
                "Error while calling WSGI application", exc_info=sys.exc_info()
            )
            log_http(environ, 500)
            raise
        finally:
            # Close the iterable if it has a close() method, per PEP 3333.
            close = getattr(iterable, "close", None)
            if callable(close):
                close()


MAX_INCOMPLETE_EVENT_SIZE = 16 * 1024


def http11_protocol(
    app: WSGIApp,
    sock: socket.socket,
    graceful_exit: threading.Event,
    *,
    url_scheme: str = "http",
    script_name: str = "",
) -> None:
    peername = sock.getpeername()[:2]
    if isinstance(peername, str):
        peername = (peername, 0)
    sockname = sock.getsockname()[:2]
    if isinstance(sockname, str):
        sockname = (sockname, 0)

    h11_connection = h11.Connection(
        our_role=h11.SERVER, max_incomplete_event_size=MAX_INCOMPLETE_EVENT_SIZE
    )
    h = H11Protocol(
        c=h11_connection,
        s=sock,
        peername=peername,
        sockname=sockname,
        graceful_exit=graceful_exit,
        url_scheme=url_scheme,
        script_name=script_name,
    )
    while not graceful_exit.is_set():
        try:
            h.call_wsgi(app)

            while True:
                event = h.get_next_event()
                match event:
                    case h11.EndOfMessage() | h11.PAUSED:
                        h.c.start_next_cycle()
                        h.response_buffer = None
                        h.header_sent = False
                        debug_logger.debug("Start next cycle in %s:%d", *h.peername)
                        break
                    case h11.Data():  # unread request body
                        pass
                    case _:
                        raise RuntimeError(f"Unexpected event: {event}")
        except ConnectionClosed:
            debug_logger.debug("Connection closed by %s:%d", *h.peername)
            break
